module env = import "../futhark_mpi/envelope"

type envelope_1d_t [n] = env.envelope_1d_t [n]
type envelope_2d_t [n][m][ty][tx] = env.envelope_2d_t [n][m][ty][tx]
type~ side_envelope_3d_t = env.side_envelope_3d_t
type~ envelope_3d_t = env.envelope_3d_t

entry get_envelope_1d [n] (xs: [n]u8) (thickness: i64) : envelope_1d_t [thickness] =
    env.get_envelope_1d xs thickness

entry get_envelope_2d [n][m] (matrix: [n][m]u8) (thickness_y: i64) (thickness_x: i64): envelope_2d_t [n][m][thickness_y][thickness_x] =
    env.get_envelope_2d matrix thickness_y thickness_x

entry get_envelope_3d [n][m][l] (cube: [n][m][l]u8) (thickness_y: i64) (thickness_x: i64) (thickness_z: i64): envelope_3d_t =
    env.get_envelope_3d cube thickness_y thickness_x thickness_z

entry augment_1d [n][tx] (xs: [n]u8) (envelope: envelope_1d_t [tx]) :[]u8 = env.augment_1d xs envelope
entry augment_2d [n][m][ty][tx] (matrix: [n][m]u8) (envelope: envelope_2d_t [n][m][tx][ty]) :[][]u8 = env.augment_2d matrix envelope
entry augment_3d [n][m][l] (cube: [n][m][l]u8) (envelope: envelope_3d_t) :[][][]u8 = env.augment_3d cube envelope

-- Program for Lattice-Boltzmann in Futhark
-- Arrays everywhere

-- Two useful functions for later
let tabulate_4d 'a  (p: i64) (n: i64) (m: i64) (o: i64) (f: i64 -> i64 -> i64 -> i64 -> a): *[p][n][m][o]a =
  map1 (f >-> tabulate_3d n m o) (iota p)

let dotprod (xs: []f32) (ys: []f32): f32 =
  reduce (+) 0 (map (\(x, y) -> x * y) (zip xs ys))

let map3d 'a [nx][ny][nz]'b (f: a -> b) (xs: [nx][ny][nz]a): [nx][ny][nz]b =
  map(\xs1 ->
    map(\xs2 ->
      map(\xs3 ->
        f xs3
      ) xs2
    ) xs1
  ) xs

let map2_3d 'a 'c [nx][ny][nz]'b (f: a -> c -> b) (xs: [nx][ny][nz]a) (ys: [nx][ny][nz]c): [nx][ny][nz]b =
  map2(\xs1 ys1 ->
    map2(\xs2 ys2 ->
      map2(\xs3 ys3 ->
        f xs3 ys3
      ) xs2 ys2
    ) xs1 ys1
  ) xs ys

------------------------------------
------------ CONSTANTS -------------
------------------------------------

module d3q27 = {

    let c: [][]f32 = [
                [ 0, 0, 0],

                [-1, 0, 0], [ 0,-1, 0], [ 0, 0,-1],
                [-1,-1, 0], [-1, 1, 0], [-1, 0,-1],
                [-1, 0, 1], [ 0,-1,-1], [ 0,-1, 1],
                [-1,-1,-1], [-1,-1, 1], [-1, 1,-1], [-1, 1, 1],

                [ 1, 0, 0], [ 0, 1, 0], [ 0, 0, 1],
                [ 1, 1, 0], [ 1,-1, 0], [ 1, 0, 1],
                [ 1, 0,-1], [ 0, 1, 1], [ 0, 1,-1],
                [ 1, 1, 1], [ 1, 1,-1], [ 1,-1, 1], [ 1,-1,-1] ]

    let w: []f32 = [
                8.0/27,
                2.0/27,  2.0/27,  2.0/27,
                1.0/54,  1.0/54,  1.0/54,
                1.0/54,  1.0/54,  1.0/54,
                1.0/216, 1.0/216, 1.0/216, 1.0/216,
                2.0/27,  2.0/27,  2.0/27,
                1.0/54,  1.0/54,  1.0/54,
                1.0/54,  1.0/54,  1.0/54,
                1.0/216, 1.0/216, 1.0/216, 1.0/216
            ]
}

let compute_rho [nx][ny][nz][q] (f: [nx][ny][nz][q]f32): [nx][ny][nz]f32 =
  map3d(\fxyz ->
    reduce (+) 0 fxyz
  ) f

let compute_j [nx][ny][nz][q][d] (f: [nx][ny][nz][q]f32)(c: [q][d]f32): [nx][ny][nz][d]f32 =
  map3d(\fxyz ->
    map(\ci ->
      dotprod ci fxyz
    ) (transpose c) -- intrinsic
  ) f

let compute_feq [nx][ny][nz][q][d] (w: [q]f32) (c: [q][d]f32)(rho: [nx][ny][nz]f32)(j: [nx][ny][nz][d]f32): [nx][ny][nz][q]f32 =
  map2_3d(\rho_xyz j_xyz ->
    let u = map(\ji -> ji / rho_xyz) j_xyz
    let u_sqr = dotprod u u

    in map2(\wi ci ->
      let c_u = dotprod ci u
      in wi * rho_xyz* (1 + 3*c_u + 4.5 * c_u * c_u - 1.5 * u_sqr)
    ) w c
  ) rho j

let collision [nx][ny][nz][q] (f: [nx][ny][nz][q]f32)(feq: [nx][ny][nz][q]f32) (omega: f32): [nx][ny][nz][q]f32 =
  map2_3d (\f_xyz feq_xyz ->
    map2(\f_i feq_i ->
      (1-omega) * f_i + omega * feq_i
    ) f_xyz feq_xyz
  ) f feq

let collide [nx][ny][nz][q][d] (f: [nx][ny][nz][q]f32) (omega: f32) (w: [q]f32) (c: [q][d]f32): [nx][ny][nz][q]f32 =
  let rho = compute_rho f
  let j = compute_j f c
  let feq = compute_feq w c rho j

  in collision f feq omega

let streaming [nx][ny][nz][q][d] (f: [nx][ny][nz][q]f32) (c: [q][d]f32): [nx][ny][nz][q]f32 =
    tabulate_4d nx ny nz q (\x y z i ->
      let next_x = (x - (i64.f32 c[i,0]) + nx ) % nx
      let next_y = (y - (i64.f32 c[i,1]) + ny ) % ny
      let next_z = (z - (i64.f32 c[i,2]) + nz ) % nz

      in f[next_x, next_y, next_z, i]
    )

let main_loop [nx][ny][nz] (fin: [nx][ny][nz][27]f32)(omega: f32) (max_t: i64): [nx][ny][nz][27]f32 =
  let f = loop f = fin for _i < max_t do
    streaming (collide f omega (copy d3q27.w) (copy d3q27.c)) (copy d3q27.c)
  in f

entry next_chunk_lbm [n][m][l][b] (chunk_lbm :[n][m][l][b]f32) :[][][][b]f32 =
    let next_lbm = (main_loop (chunk_lbm :> [n][m][l][27]f32) 1.0 1) :> [][][][b]f32
    in next_lbm[1:n-1, 1:m-1, 1:l-1, :]
